@@ -12,4 +12,16 @@ py::object cpp_katz_centrality(
     py::object py_max_iter,
     py::object py_tol,
     py::object py_normalized
+);
+
+py::object degree_centrality(py::object G);
+py::object in_degree_centrality(py::object G);
+py::object out_degree_centrality(py::object G);
+
+py::object cpp_eigenvector_centrality(
+    py::object G,
+    py::object py_max_iter,
+    py::object py_tol,
+    py::object py_nstart,
+    py::object py_weight
 );
\ No newline at end of file
